import csv
import numpy as np
from scipy.signal import gaussian
from sklearn.preprocessing import normalize
import wfdb

personal_data_indices = [200, 201, 202, 203, 205, 207, 208, 209, 210, 212,
                         213, 214, 215, 219, 220, 221, 222, 223, 228, 230,
                         231, 232, 233, 234]

arrhythmia_dict = {1: 1,    # Normal beat
                   2: 2,    # LBBB
                   3: 3,    # RBBB
                   4: 4,    # Aberrated atrial premature beat
                   8: 4,    # Atrial premature contraction
                   5: 5,    # Premature ventricular contraction
                   10: 5,   # ventricular escape beat
                   6: 6,    # Fusion of ventricular and normal beat
                   38: 7,
                   12: 7,
                   13: 7
                   }


def get_receptive_fields(signal, step, overlap, sigma= 5):
    total = create_on_off_wavelet(signal[:,0:step])
    for i in range(1,signal.shape[1]//overlap - 1):
        on_off_signal = create_on_off_wavelet(signal[:,overlap*i:overlap*i+step])
        total = np.concatenate((total, on_off_signal),axis=1)
    return total


def create_on_off_wavelet(wavelet_matrix):
    on_matrix = wavelet_matrix * (wavelet_matrix >= 0)
    off_matrix = -wavelet_matrix * (wavelet_matrix < 0)
    return np.concatenate((on_matrix, off_matrix), axis=1)


def read_personal_data(person_num, binary_class=False, balance=None, data_type='train', root='input_data/'):
    signal = []
    label = []
    if data_type == 'train':
        # personal data in train
        filename = root + 'Ptrain/ECG_data_' + str(person_num) + '_1_train.csv'
        file_obj = open(filename)
        reader = csv.reader(file_obj)
        for row in reader:
            beat = list(map(float, row[0:250])) + 6 * [0]
            signal.append(beat)
            if binary_class:
                label.append(0 if int(float(row[-1])) == 1 else 1)
            else:
                label.append(int(float(row[-1]))-1)
        file_obj.close()
        if balance is not None:
            raw_signal = list.copy(signal)
            raw_label = list.copy(label)
            main_len = len(np.where(np.array(label) == 0)[0]) + \
                len(np.where(np.array(label) == 1)[0]) + len(np.where(np.array(label) == 2)[0])
            for arrhythmia_label in balance:
                arrhythmia_len = len(np.where(np.array(label) == arrhythmia_label)[0])
                print(arrhythmia_len)
                if arrhythmia_len == 0:
                    continue
                for i, row in enumerate(raw_signal):
                    if raw_label[i] == arrhythmia_label:
                        for repeat in range(main_len//arrhythmia_len):
                            rand_index = np.random.randint(len(label))
                            signal.insert(rand_index, row)
                            label.insert(rand_index, arrhythmia_label)
    else:
        filename = root + 'Ptest/ECG_data_' + str(person_num) + '_1_test.csv'
        file_obj = open(filename)
        reader = csv.reader(file_obj)
        for row in reader:
            beat = list(map(float, row[0:250])) + 6 * [0]
            normalized_beat = normalize(np.array(beat).reshape(-1, 1), axis=0)
            signal.append(normalized_beat)

            if binary_class:
                label.append(0 if int(float(row[-1])) == 1 else 1)
            else:
                label.append(int(float(row[-1]))-1)

    return np.array(signal).squeeze()[:], np.array(label)


def read_data(model_num, binary_class=False, balance=None, data_type='train'):
    record = wfdb.rdrecord('../data/raw/' + str(model_num), sampfrom=0, sampto=None,
                           channels=[0])
    ann = wfdb.rdann('../data/raw/' + str(model_num), 'atr', sampfrom=0, sampto=None,
                     return_label_elements=['label_store'])

    signal = []
    label = []
    if data_type == 'train':
        # personal data in train
        for index, r_peak in enumerate(ann.sample):
            if r_peak > 360 * 60 * 5:
                break
            if ann.label_store[index] not in arrhythmia_dict:
                continue
            arrhythmia = ann.label_store[index]
            try:
                beat = record.p_signal[r_peak - 100:r_peak + 150, 0]
                if beat.shape[0] != 250:
                    continue
                beat = np.concatenate((beat,np.zeros(6)))
                signal.append(beat)
                if binary_class:
                    label.append(0 if arrhythmia_dict[arrhythmia] == 1 else 1)
                else:
                    label.append(arrhythmia_dict[arrhythmia] - 1)
            except ValueError:
                print('error in', index)

        if balance is not None:
            raw_signal = list.copy(signal)
            raw_label = list.copy(label)
            main_len = len(np.where(np.array(label) == 0)[0]) + \
                       len(np.where(np.array(label) == 1)[0]) + len(np.where(np.array(label) == 2)[0])
            for arrhythmia_label in balance:
                arrhythmia_len = len(np.where(np.array(label) == arrhythmia_label)[0])
                print(arrhythmia_len)
                if arrhythmia_len == 0:
                    continue
                for i, row in enumerate(raw_signal):
                    if raw_label[i] == arrhythmia_label:
                        for repeat in range(main_len // arrhythmia_len):
                            rand_index = np.random.randint(len(label))
                            signal.insert(rand_index, row)
                            label.insert(rand_index, arrhythmia_label)
    else:
        for index, r_peak in enumerate(ann.sample):
            if r_peak <= 360 * 60 * 5:
                continue
            if ann.label_store[index] not in arrhythmia_dict:
                continue
            arrhythmia = ann.label_store[index]
            try:
                beat = record.p_signal[r_peak - 100:r_peak + 150, 0]
                if beat.shape[0] != 250:
                    continue
                beat = np.concatenate((beat, np.zeros(6)))
                signal.append(beat)
                if binary_class:
                    label.append(0 if arrhythmia_dict[arrhythmia] == 1 else 1)
                else:
                    label.append(arrhythmia_dict[arrhythmia] - 1)
            except ValueError:
                print('error in', index)

    return np.array(signal).squeeze(), np.array(label)
