# Please refer to README.md

import json
import os.path
import argparse
import brian2 as b2
import numpy as np
from sklearn.metrics import confusion_matrix
from sklearn.neighbors import KNeighborsClassifier
from sklearn.preprocessing import normalize
from scipy.signal import gaussian
import logging

from data import read_data, get_receptive_fields, read_personal_data

# Initialize the parameters
PATIENT_NUM = 200
TEST_LEN = 2168
TRAIN_LEN = 690

COEFF_NUM = 448 * 2  # Based on Q
PROGRESS_NUM = 100
DECAY_RATE = 0.05
LR_INIT = 0.05
N_window = 14   # 2 * Q
DURATION_BIN = 200 * b2.ms
N_per_field = 10
N_S = N_window * N_per_field  # Number of neurons in the STDP layer
STEP_RECEPTIVE = 64  # Number of input neurons in a window
OVERLAP_RECEPTIVE = 32
NORMAL_AMP = 1500
NORMAL_BIAS = 5
NORMAL_INTERVAL = 10  # Interval for normalizing the weights
R_TARGET = 20  # The average rate of spikes in the Gaussian layer
W_MAX_EE = 8  # Maximum weight in the STDP layer
STDP_THRESHOLD = 12.0
N_out = 4  # Number of neurons in the RSTDP layer
RSTDP_EPOCH_NUM =  400
RSTDP_DECAY_RATE = 0.997
N_R = 4  # Number of classes in the RSTDP layer
BATCH_SIZE = 16


def phase_input (string):
    if not string in ['train', 'test']:
        raise argparse.ArgumentTypeError('Phase has to be \'train\' for retraining or \'test\' for inference')
    return string


parser = argparse.ArgumentParser()
parser.add_argument('--P', type=phase_input, default='test')      # Phase: test or train?
args = parser.parse_args()
mode = args.P

# Loading data 
if mode == 'train':
    input_data = json.load(open('input_data/spikes_train_'+str(PATIENT_NUM)+'.json', 'r'))
else:
    input_data = json.load(open('input_data/spikes_test_'+str(PATIENT_NUM)+'.json', 'r'))

# Spike trains contain neuron indices, spike times and the labels
input_indices = input_data['indices']
input_times = input_data['times'] * b2.ms
y_input = input_data['labels']

# Define models for the neurons and synapses
hidden_neurons_eqs = '''
dv/dt = -v/(10*ms) : 1 (unless refractory)
'''

eqs_stdp = '''
w                       : 1
pre : second
post : second
freeze : 1
'''

eqs_stdp_pre_ee = '''
pre = t
dpost = post - pre
isInterval = (floor(t/DURATION_BIN) == floor(post/DURATION_BIN))
dw = clip((-1 - dpost /T)*(1+(neg_r*w/(neg_r*w+wmax_ee))), -neg_r , 0) * isInterval
w = clip(w + (learning_rate * dw * isTrainable), 0, 100)
v += w
post = 0*ms
'''
eqs_stdp_post_ee = '''
post = t
dpre = post - pre
isInterval = (floor(t/DURATION_BIN) == floor(pre/DURATION_BIN))
dw = clip(1 - dpre / T, 0 ,1) * isInterval
w = w + (learning_rate * dw * isTrainable)
pre = 0*ms
'''

eqs_inhibitory = '''
w : 1
'''

eqs_stdp_pre_ei = '''
v += 1
'''

eqs_inhibitory_pre = '''
pre = t
dpost = post - pre
isInterval = (floor(t/DURATION_BIN) == floor(post/DURATION_BIN))
isNear = (dpost >= -1*ms) 
dw = (-0.1 + isNear * 1/(1+neg_inh*w)) * isTrainable * isInterval * freeze
w = clip(w + (learning_rate * dw), 0, wmax_ie)
v -= (w * freeze)
'''

eqs_inhibitory_post = '''
post = t
dpre = post - pre
isInterval = (floor(t/DURATION_BIN) == floor(pre/DURATION_BIN))
isNear = (dpre <= 0.5*ms) * isInterval
dw = (-0.1 + isNear * 1/(1+neg_inh*w)) * isTrainable * isInterval * freeze
w = clip(w + (learning_rate * dw), 0, wmax_ie)
'''

eqs_rstdp_ee = '''
w : 1
'''

eqs_rstdp_pre = '''
v += w
'''

# Initialize the network parameters
connection_params = {'learning_rate': LR_INIT,
                     'wmax_ee': W_MAX_EE,
                     'T': 4 * b2.ms,
                     'neg_r': 1.0,
                     'isTrainable': 1,
                     'wmax_ie': 100,
                     'neg_inh': 1, }

hidden_neuron_params = {'threshold': STDP_THRESHOLD,
                        'refractory': 1 * b2.ms}

layer1_neuron_params = {'threshold': 0.95,
                        'refractory': 0.5 * b2.ms}

r_connection_params = {'a_r+': 0.002,
                       'a_r-': -0.004,
                       'a_p+': 0,
                       'a_p-': -0.005,
                       'factor': 1,
                       'w_max': 1.0,
                       'max_negative_rate': 3}

r_neuron_params = {'r_th': 12.0,
                   'refractory': 1 * b2.ms}

# Creating neurons and synapses
# Define the neurons as spike generator of a simple neuron group
input_neurons = b2.SpikeGeneratorGroup(N=COEFF_NUM, indices=input_indices, times= input_times, period=len(y_input) * DURATION_BIN)

gaussian_neurons = b2.NeuronGroup(N=COEFF_NUM, model=hidden_neurons_eqs, method='exact',
                                  threshold='v>threshold', reset='v-=threshold', refractory=1 * b2.ms,
                                  namespace=layer1_neuron_params, dt=1 * b2.ms)

STDP_neurons = b2.NeuronGroup(N=N_S, model=hidden_neurons_eqs, method='exact',
                              threshold='v>threshold', reset='v = 0', refractory='refractory',
                              dt=1 * b2.ms, namespace=hidden_neuron_params)

inhibitory_neurons = b2.NeuronGroup(N=N_S, model=hidden_neurons_eqs, method='exact',
                                    threshold='v>0.9', reset='v = 0', refractory=1 * b2.ms,
                                    dt=1 * b2.ms)

r_neurons = b2.NeuronGroup(N=N_R, model=hidden_neurons_eqs, method='exact',
                           threshold='v>r_th', reset='v = 0', refractory='refractory',
                           dt=1 * b2.ms, namespace=r_neuron_params)

# Define the synapses which connect the neuron layers
gaussian_synapse = b2.Synapses(input_neurons, gaussian_neurons, model='w:1',
                               on_pre='v += w', dt=1 * b2.ms)
gaussian_synapse.connect(j='i')

s_synapses = b2.Synapses(gaussian_neurons, STDP_neurons, model=eqs_stdp, on_pre=eqs_stdp_pre_ee,
                         on_post=eqs_stdp_post_ee, namespace=connection_params, dt=1 * b2.ms)
s_synapses.connect('floor((N_window*j)/N_S)==floor((N_window*i)/COEFF_NUM)')

synapses_ei = b2.Synapses(STDP_neurons, inhibitory_neurons, model=eqs_inhibitory,
                          on_pre=eqs_stdp_pre_ei, dt=1 * b2.ms)
synapses_ei.connect(j='i')

inhibitory_ie = b2.Synapses(inhibitory_neurons, STDP_neurons, model=eqs_stdp,
                            on_pre=eqs_inhibitory_pre, on_post=eqs_inhibitory_post,
                            namespace=connection_params, dt=1 * b2.ms)
inhibitory_ie.connect('j!=i and floor((N_window*j)/N_S)==floor((N_window*i)/N_S)')

r_synapses = b2.Synapses(STDP_neurons, r_neurons, model=eqs_rstdp_ee, on_pre=eqs_rstdp_pre, dt=1 * b2.ms)
r_synapses.connect(True)


def prepare_labels(y_input):
    labels = np.zeros_like(y_input)
    for i in range(len(y_input)):
        if y_input[i] in [0, 1, 2, 3]:  # Normal, LBB, RBB, S
            labels[i] = 0
        elif y_input[i] == 4:  # PVC
            labels[i] = 1
        elif y_input[i] == 5:  # F
            labels[i] = 2
        elif y_input[i] == 6:  # Q
            labels[i] = 3
        else:
            print("unknown label ", y_input[i])
    return labels


def print_progress_bar(iteration, total, prefix='', suffix='', decimals=1, length=100, fill='█'):
    if iteration % (total // length) != 0:
        return
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix))
    if iteration == total:
        print()


def index_to_label(index):
    return index % N_out


def update_weight_reward(winner_index, winner_time, true_label,
                         input_spike_times, input_spike_i, hit_miss_factor):
    # This method is used in training the RSTDP layer. When a reward is received, the weights will change
    delta_w = np.zeros(N_S)
    first_spike_time = np.ones(N_S) * -1
    for array_index, neuron_index in enumerate(input_spike_i):
        if first_spike_time[neuron_index] == -1:
            first_spike_time[neuron_index] = input_spike_times[array_index]

    for neuron_index in range(N_S):
        if first_spike_time[neuron_index] == -1:
            delta_w[neuron_index] = r_connection_params['a_r-']
        elif first_spike_time[neuron_index] < winner_time:
            delta_w[neuron_index] = r_connection_params['a_r+']
        elif first_spike_time[neuron_index] > winner_time:
            delta_w[neuron_index] = r_connection_params['a_r-']

    delta_w *= hit_miss_factor

    if true_label == 1:
        delta_w *= r_connection_params['factor']

    w_reshaped = np.reshape(r_synapses.w, (N_S, N_R))
    w_reshaped_winner = w_reshaped[:, winner_index]
    w_max = r_connection_params['w_max']
    delta_w *= (w_reshaped_winner*(w_max - w_reshaped_winner))
    w_reshaped[:, winner_index] += delta_w


def update_weight_punishment(winner_index, winner_time, true_label,
                             input_spike_times, input_spike_i, hit_miss_factor):
    # This method is used in training the RSTDP layer. When a punishment is received, the weights will change
    delta_w = np.zeros(N_S)
    first_spike_time = np.ones(N_S) * -1
    for array_index, neuron_index in enumerate(input_spike_i):
        if first_spike_time[neuron_index] == -1:
            first_spike_time[neuron_index] = input_spike_times[array_index]

    for neuron_index in range(N_S):
        if first_spike_time[neuron_index] < winner_time and first_spike_time[neuron_index] != -1:
            delta_w[neuron_index] = r_connection_params['a_p-']

    delta_w *= hit_miss_factor

    if true_label == 1 :
        delta_w *= r_connection_params['factor']

    w_reshaped = np.reshape(r_synapses.w, (N_S, N_R))
    w_max = r_connection_params['w_max']
    w_reshaped_winner = w_reshaped[:, winner_index]
    delta_w *= (w_reshaped_winner * (w_max - w_reshaped_winner))
    w_reshaped[:, winner_index] += delta_w


def get_pointer_list(spike_time, start_time, size, initial_count):
    pointers = np.zeros(size + 1, dtype=np.int)
    end_time = start_time + DURATION_BIN * (size + 1)
    pointers[0] = initial_count
    current_pos = 1
    for index, t in enumerate(np.append(spike_time, end_time)):
        while t * b2.second > (start_time + current_pos * DURATION_BIN):
            pointers[current_pos] = index
            current_pos += 1
    return pointers


def init_weights(mode='train'):
    # Load Gaussian weights
    if mode is 'trian':
        gaussian_file_name = 'models/gaussian_model' + str(PATIENT_NUM) + '.json'
        data = json.load(open(gaussian_file_name, 'r'))
        init_g = np.array(data['weights'], dtype=np.float)
    else:
        init_g = np.tile(0.5 * gaussian(STEP_RECEPTIVE, 15), N_window)
    gaussian_synapse.w = init_g

    # Load STDP weights
    if mode is 'train':
        stdp_w = np.random.normal(W_MAX_EE * 0.8, 0.5, N_S * STEP_RECEPTIVE)
    else:
        stdp_file_name = 'models/STDP_model' + str(PATIENT_NUM) + '.json'
        data = json.load(open(stdp_file_name, 'r'))
        stdp_w = data['weights']
    s_synapses.w = stdp_w

    # Initialize inhibitory weights
    if mode is 'train':
        inhibitory_ie.w = np.abs(
            np.random.normal(W_MAX_EE * 0.2, 0.5, N_S * (N_S // N_window - 1)))
        inhibitory_ie.freeze = np.random.binomial(1, 0.75, N_S * (N_S // N_window - 1))

    # Load RSTDP weights
    if mode is 'train':
        epsilon = 0.1
        init_w = np.random.normal(r_connection_params['w_max'] * 0.6, 0.25, N_S * N_R)
        rstdp_w = np.clip(init_w, epsilon, r_connection_params['w_max'] - epsilon)
    else:
        file_name = 'models/RSTDP_model' + str(PATIENT_NUM) + '.json'
        data = json.load(open(file_name, 'r'))
        rstdp_w = data['weights']
    r_synapses.w = rstdp_w


def get_STDP_accuracy(net,  x, y, x_test, y_test, spikemon):
    print('Getting STDP accuracy ...')
    connection_params['isTrainable'] = 0
    stimulus = b2.TimedArray(np.concatenate((x, x_test), axis=0) * b2.Hz, dt=DURATION_BIN)
    hidden_count = np.zeros((x.shape[0], N_S), dtype=np.int)
    net.run(DURATION_BIN * len(x))
    for neuron_i, neuron_t in zip(spikemon.i[:], (spikemon.t[:]) // DURATION_BIN):
        hidden_count[int(neuron_t)][neuron_i] += 1
    train_spikes = len(spikemon.t[:])
    neigh = KNeighborsClassifier(n_neighbors=3)
    neigh.fit(hidden_count, y)

    counter = np.zeros((x_test.shape[0], N_S), dtype=np.int)
    base_time = net.t
    net.run(DURATION_BIN * len(x_test))
    for neuron_i, neuron_t in zip(spikemon.i[train_spikes:], (spikemon.t[train_spikes:] - base_time) // DURATION_BIN):
        counter[int(neuron_t)][neuron_i] += 1
    predicted = neigh.predict(counter)
    print('confusion matrix : \t', confusion_matrix(y_test, predicted))
    result = {'results': confusion_matrix(y_test, predicted)[:].tolist(),
              'weights': s_synapses.w[:].tolist(),
              }
    json.dump(result, open('models/STDP_model' + str(PATIENT_NUM) + '.json', 'w'))


def get_accuracy(net, y_test, r_spikemon):
    print('Getting accuracy ...')
    connection_params['isTrainable'] = 0
    start_time = net.t
    net.run(DURATION_BIN * TEST_LEN)
    hidden_pointer = get_pointer_list(r_spikemon.t[:], start_time, TEST_LEN, 0)
    test_confusion = np.zeros((N_out, N_out), dtype=np.int)

    out_index = np.zeros(TEST_LEN)
    for data_index in range(TEST_LEN):
        if hidden_pointer[data_index] == hidden_pointer[data_index + 1]:
            out_index[data_index] = -1
            try:
                test_confusion[y_test[data_index], 0] += 1
            except IndexError as e:
                logging.info("%s in index: %d" % (e, data_index))
            continue
        start_point = hidden_pointer[data_index]
        winner_index = r_spikemon.i[start_point]
        predicted = index_to_label(winner_index)
        out_index[data_index] = winner_index
        try:
            test_confusion[y_test[data_index], predicted] += 1
        except IndexError as e:
            logging.info("%s in index: %d" % (e, data_index))
    print('test confusion matrix : \n', test_confusion)
    result = {'results': test_confusion.tolist(),
              'weights': r_synapses.w[:].tolist()}
    json.dump(result, open('outputs/TEST_model' + str(PATIENT_NUM) + '.json', 'w'))


def train_gaussian(net, g_spikemon):
    print('Training the Gaussian layer ...')
    connection_params['isTrainable'] = 0
    g = [1] * N_window
    for progress in range(PROGRESS_NUM):
        for num in range(TRAIN_LEN // BATCH_SIZE):
            previous_count = np.copy(g_spikemon.count[:])
            net.run(DURATION_BIN * BATCH_SIZE)
            current_count = np.copy(g_spikemon.count[:]) - previous_count
            for field in range(N_window):
                cnt = current_count[field * STEP_RECEPTIVE:(field + 1) * STEP_RECEPTIVE]
                mean_cnt = np.mean(cnt)
                T = DURATION_BIN / b2.second
                p = (mean_cnt / (T * BATCH_SIZE)) / R_TARGET
                delta_g = 1 - p
                g[field] += delta_g
                new_g = g[field] * 0.5 * gaussian(STEP_RECEPTIVE, 15)
                gaussian_synapse.w[field * STEP_RECEPTIVE:(field + 1) * STEP_RECEPTIVE] = np.clip(new_g, 0, 1)


def train_STDP(net):
    print('Training the STDP layer ...')
    previous_w = (np.copy(gaussian_synapse.w), np.copy(s_synapses.w),
                  np.copy(inhibitory_ie.w), np.copy(r_synapses.w))
    connection_params['isTrainable'] = 1
    for progress in range(PROGRESS_NUM):
        net.restore(name='network')
        gaussian_synapse.w, s_synapses.w, inhibitory_ie.w, r_synapses.w = previous_w
        inhibitory_ie.freeze = np.random.binomial(1, 0.75, N_S * (N_S // N_window - 1))
        net.run(DURATION_BIN * TRAIN_LEN)
        print_progress_bar(progress, PROGRESS_NUM,
                         prefix='train data : ', length=PROGRESS_NUM // 10)
        connection_params['learning_rate'] = LR_INIT * np.exp(-DECAY_RATE * progress)
        previous_w = (np.copy(gaussian_synapse.w), np.copy(s_synapses.w),
                      np.copy(inhibitory_ie.w), np.copy(r_synapses.w))


def train_RSTDP(net, y_train, s_spikemon, r_spikemon):
    print('Training the RSTDP layer ...')
    connection_params['isTrainable'] = 0
    previous_w = (np.copy(gaussian_synapse.w), np.copy(s_synapses.w),
                  np.copy(inhibitory_ie.w), np.copy(r_synapses.w))
    for epoch in range(RSTDP_EPOCH_NUM):
        net.restore(name='network')
        gaussian_synapse.w, s_synapses.w, inhibitory_ie.w, r_synapses.w = previous_w
        for batch_index in range(TRAIN_LEN // BATCH_SIZE):
            start_time = net.t
            r_previous_count = len(r_spikemon.t[:])
            s_previous_count = len(s_spikemon.t[:])
            net.run(DURATION_BIN * BATCH_SIZE)
            input_pointer = get_pointer_list(s_spikemon.t[:], start_time, BATCH_SIZE, s_previous_count)
            hidden_pointer = get_pointer_list(r_spikemon.t[:], start_time, BATCH_SIZE, r_previous_count)
            for data_index in range(BATCH_SIZE):
                train_index = batch_index * BATCH_SIZE + data_index
                start_point = hidden_pointer[data_index]
                input_start_point = input_pointer[data_index]
                input_end_point = input_pointer[data_index + 1]

                if hidden_pointer[data_index] == hidden_pointer[data_index + 1]:
                    winner_time = (start_time + (data_index + 1) * DURATION_BIN) / b2.ms
                    for index in range(N_R):
                        update_weight_reward(index, winner_time, y_train[train_index],
                                             s_spikemon.t[input_start_point:input_end_point] / b2.ms,
                                             s_spikemon.i[input_start_point:input_end_point], 1)
                    continue

                winner_list = []
                winner_time = r_spikemon.t[start_point] / b2.ms
                offset = 0
                while r_spikemon.t[start_point+offset] / b2.ms == winner_time:
                    winner_list.append(r_spikemon.i[start_point+offset])
                    offset += 1
                    if start_point+offset >= len(r_spikemon.t[:]):
                        break
                for winner_index in winner_list:
                    if index_to_label(winner_index) == y_train[train_index]:
                        update_weight_reward(winner_index, winner_time, y_train[train_index],
                                             s_spikemon.t[input_start_point:input_end_point] / b2.ms,
                                             s_spikemon.i[input_start_point:input_end_point], 1)
                    else:
                        update_weight_punishment(winner_index, winner_time, y_train[train_index],
                                                 s_spikemon.t[input_start_point:input_end_point] / b2.ms,
                                                 s_spikemon.i[input_start_point:input_end_point], 1)
        previous_w = (np.copy(gaussian_synapse.w), np.copy(s_synapses.w),
                      np.copy(inhibitory_ie.w), np.copy(r_synapses.w))
        r_connection_params['a_r+'] *= DECAY_RATE
        r_connection_params['a_r-'] *= DECAY_RATE
        r_connection_params['a_p+'] *= DECAY_RATE
        r_connection_params['a_p-'] *= DECAY_RATE
        print_progress_bar(epoch, RSTDP_EPOCH_NUM, prefix='train epoch %d : ' % epoch, length=RSTDP_EPOCH_NUM//10)


def main(mode='train'):
    print("Preparing network ...")
    s_spikemon = b2.SpikeMonitor(STDP_neurons, record=True)
    r_spikemon = b2.SpikeMonitor(r_neurons, record=True)
    g_spikemon = b2.SpikeMonitor(gaussian_neurons, record=False)  # we want just the number of spikes

    @b2.network_operation(dt=DURATION_BIN * NORMAL_INTERVAL)
    def normalize_weight():
        if connection_params['isTrainable'] == 0:
            return
        w_1D = np.zeros_like(s_synapses.w)
        w_copy = np.copy(s_synapses.w)
        w_reshaped = np.reshape(w_copy, (COEFF_NUM, N_per_field))
        for i in range(N_window):
            w_norm = np.linalg.norm(w_reshaped[i * STEP_RECEPTIVE:(i + 1) * STEP_RECEPTIVE, :], axis=0)
            w_normal = w_reshaped[i * STEP_RECEPTIVE:(i + 1) * STEP_RECEPTIVE, :] / w_norm * W_MAX_EE
            w_1D[STEP_RECEPTIVE * N_per_field * i:STEP_RECEPTIVE * N_per_field * (i + 1)] = np.reshape(w_normal, (-1,))
        s_synapses.w = np.reshape(w_1D, (-1,))

    @b2.network_operation(dt=DURATION_BIN)
    def reset_neurons():
        STDP_neurons.v = 0
        inhibitory_neurons.v = 0
        gaussian_neurons.v = 0
        r_neurons.v = 0

    net = b2.Network(b2.collect())
    net.store(name='network')
    print('Running model ', PATIENT_NUM)

    print('Loading data ...')
    y = prepare_labels(y_input)

    if mode == 'test':
        if not os.path.exists('models/gaussian_model' + str(PATIENT_NUM) + '.json'):
            print('Gaussian model not found')
        elif not os.path.exists('models/STDP_model' + str(PATIENT_NUM) + '.json'):
            print('STDP model not found')
        elif not os.path.exists('models/RSTDP_model' + str(PATIENT_NUM) + '.json'):
            print('RSTDP model not found')
        else:
            init_weights(mode=mode)
            get_accuracy(net, y, r_spikemon)
            return

    print('Start training ...')
    init_weights(mode='train')
    train_gaussian(net, g_spikemon)
    train_STDP(net)
    train_RSTDP(net, y, s_spikemon, r_spikemon)


if __name__ == '__main__':
    main(mode)
